import os

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.invenio.invenio_api_views import InvenioAPIViewsBuilder
from oarepo_model_builder.invenio.invenio_config import InvenioConfigBuilder
from oarepo_model_builder.invenio.invenio_ext import InvenioExtBuilder
from oarepo_model_builder.invenio.invenio_proxies import InvenioProxiesBuilder
from oarepo_model_builder.invenio.invenio_record import InvenioRecordBuilder
from oarepo_model_builder.invenio.invenio_record_dumper import (
    InvenioRecordDumperBuilder,
)
from oarepo_model_builder.invenio.invenio_record_metadata import (
    InvenioRecordMetadataBuilder,
)
from oarepo_model_builder.invenio.invenio_record_permissions import (
    InvenioRecordPermissionsBuilder,
)
from oarepo_model_builder.invenio.invenio_record_pid_provider import (
    InvenioRecordPIDProviderBuilder,
)
from oarepo_model_builder.invenio.invenio_record_resource import (
    InvenioRecordResourceBuilder,
)
from oarepo_model_builder.invenio.invenio_record_resource_config import (
    InvenioRecordResourceConfigBuilder,
)
from oarepo_model_builder.invenio.invenio_record_service import (
    InvenioRecordServiceBuilder,
)
from oarepo_model_builder.invenio.invenio_record_service_config import (
    InvenioRecordServiceConfigBuilder,
)
from oarepo_model_builder.invenio.invenio_record_ui_serializer import (
    InvenioRecordUISerializerBuilder,
)
from oarepo_model_builder.invenio.invenio_version import InvenioVersionBuilder
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.schema import ModelSchema

from .utils import strip_whitespaces


def build_python_model(model, output_builders, fn):
    builder = ModelBuilder(
        output_builders=output_builders,
        outputs=[PythonOutput],
        filesystem=InMemoryFileSystem(),
    )
    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {
                        "use-isort": False,
                        "use-black": False,
                        "use-autoflake": False,
                    },
                    "opensearch": {"version": "os-v2"},
                },
                "record": {"module": {"qualified": "test"}, **model},
            },
        ),
        profile="record",
        model_path=["record"],
        output_dir="",
    )

    return builder.filesystem.read(fn)


def test_record_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordBuilder],
        os.path.join("test", "records", "api.py"),  # NOSONAR
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
from invenio_records.systemfields import ConstantField
from invenio_records_resources.records.systemfields import IndexField
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2
from invenio_records_resources.records.systemfields.pid import PIDField
from invenio_records_resources.records.systemfields.pid import PIDFieldContext
from test.records.models import TestMetadata
from test.records.dumper import TestDumper
from invenio_records_resources.records.api import Record as InvenioRecord
class TestRecord(InvenioRecord):
    model_cls = TestMetadata
    schema = ConstantField("$schema", "local://test-1.0.0.json")
    index = IndexField("test-test-1.0.0")
    pid = PIDField(
        provider=TestIdProvider,
        context_cls=PIDFieldContext,
        create=True
    
    )
    dumper_extensions = []
    dumper = TestDumper(extensions=dumper_extensions)
"""
    )


def test_record_pid_provider_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [
            InvenioRecordPIDProviderBuilder,
            InvenioRecordBuilder,
        ],
        os.path.join("test", "records", "api.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2
from invenio_records_resources.records.systemfields.pid import PIDField
from invenio_records_resources.records.systemfields.pid import PIDFieldContext
from invenio_records.systemfields import ConstantField
from invenio_records_resources.records.systemfields import IndexField
from test.records.models import TestMetadata
from test.records.dumper import TestDumper
from invenio_records_resources.records.api import Record as InvenioRecord

class TestIdProvider(RecordIdProviderV2):
    pid_type = "test"

class TestRecord(InvenioRecord):
    model_cls = TestMetadata
    schema = ConstantField("$schema", "local://test-1.0.0.json")
    index = IndexField("test-test-1.0.0")
    pid = PIDField(
        provider=TestIdProvider,
        context_cls=PIDFieldContext,
        create=True
    
    )
    dumper_extensions = []
    dumper = TestDumper(extensions=dumper_extensions)
"""
    )


def test_record_metadata_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [
            InvenioRecordMetadataBuilder,
        ],
        os.path.join("test", "records", "models.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from invenio_db import db


from invenio_records.models import RecordMetadataBase


class TestMetadata(db.Model, RecordMetadataBase):
    """Model for TestRecord metadata."""

    __tablename__ = "test_metadata"

    # Enables SQLAlchemy-Continuum versioning
    __versioned__ = {}
'''
    )


def test_ext_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioExtBuilder],
        os.path.join("test", "ext.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
import re
from test import config as config


class TestExt:
    def __init__(self, app=None):
        """Extension initialization."""
        self.resource = None
        self.service = None
        
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        
        self.init_config(app)
        if not self.is_inherited():
            self.init_resource(app)
            self.register_flask_extension(app)

    def register_flask_extension(self, app):
        
        app.extensions["test"] = self

    def init_resource(self, app):
        """Initialize vocabulary resources."""
        self.service = app.config["TEST_RECORD_SERVICE_CLASS"](
            config=app.config["TEST_RECORD_SERVICE_CONFIG"](),
        )
        self.resource = app.config["TEST_RECORD_RESOURCE_CLASS"](
            service=self.service,
            config=app.config["TEST_RECORD_RESOURCE_CONFIG"](),
        )

    def init_config(self, app):
        """Initialize configuration."""
        for identifier in dir(config):
            if re.match('^[A-Z_0-9]*$', identifier) and not identifier.startswith('_'):
                app.config.setdefault(identifier, getattr(config, identifier))

    def is_inherited(self):
        from importlib_metadata import entry_points

        ext_class = type(self)
        for ep in entry_points(group='invenio_base.apps'):
            loaded = ep.load()
            if loaded is not ext_class and issubclass(ext_class, loaded):
                return True
        for ep in entry_points(group='invenio_base.api_apps'):
            loaded = ep.load()
            if loaded is not ext_class and issubclass(ext_class, loaded):
                return True
        return False
'''
    )


def test_proxies_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioProxiesBuilder],
        os.path.join("test", "proxies.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from flask import current_app
from werkzeug.local import LocalProxy

def _ext_proxy(attr):
    return LocalProxy(
        lambda: getattr(current_app.extensions["test"], attr))

current_service = _ext_proxy('service')
"""Proxy to the instantiated vocabulary service."""

current_resource = _ext_proxy('resource')
"""Proxy to the instantiated vocabulary resource."""
        '''
    )


def test_config_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioConfigBuilder],
        os.path.join("test", "config.py"),  # NOSONAR
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
from test.services.records.config import TestServiceConfig
from test.services.records.service import TestService
from test.resources.records.config import TestResourceConfig
from test.resources.records.resource import TestResource


TEST_RECORD_RESOURCE_CONFIG = TestResourceConfig
TEST_RECORD_RESOURCE_CLASS = TestResource
TEST_RECORD_SERVICE_CONFIG = TestServiceConfig
TEST_RECORD_SERVICE_CLASS = TestService
        """
    )


def test_version_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioVersionBuilder],
        os.path.join("test", "version.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
__version__ = "1.0.0"
        """
    )


def test_resource_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordResourceBuilder],
        os.path.join("test", "resources", "records", "resource.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from invenio_records_resources.resources import RecordResource
class TestResource(RecordResource):
    """TestRecord resource."""
    # here you can for example redefine
    # create_url_rules function to add your own rules
        '''
    )


def test_resource_config_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordResourceConfigBuilder],
        os.path.join("test", "resources", "records", "config.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
import importlib_metadata
from flask_resources import ResponseHandler

from test.resources.records.ui import TestUIJSONSerializer
from invenio_records_resources.resources import RecordResourceConfig

class TestResourceConfig(RecordResourceConfig):
    """TestRecord resource config."""

    blueprint_name = 'test'
    url_prefix = '/test/'

    @property
    def response_handlers(self):
        entrypoint_response_handlers = {}
        for x in importlib_metadata.entry_points(group='invenio.test.response_handlers'):
            entrypoint_response_handlers.update(x.load())
        return {
            "application/vnd.inveniordm.v1+json": ResponseHandler(TestUIJSONSerializer()),
            **super().response_handlers,
            **entrypoint_response_handlers
        }
        '''
    )


def test_service_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordServiceBuilder],
        os.path.join("test", "services", "records", "service.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from invenio_records_resources.services import RecordService as InvenioRecordService

class TestService(InvenioRecordService):
    """TestRecord service."""        
'''
    )


def test_service_config():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordServiceConfigBuilder],
        os.path.join("test", "services", "records", "config.py"),
    )
    print(data)
    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from invenio_records_resources.services import RecordServiceConfig as InvenioRecordServiceConfig
from oarepo_runtime.config.service import PermissionsPresetsConfigMixin
from invenio_records_resources.services import RecordLink
from invenio_records_resources.services import pagination_links
from test.records.api import TestRecord
from test.services.records.permissions import TestPermissionPolicy
from test.services.records.schema import TestSchema
from test.services.records.search import TestSearchOptions


class TestServiceConfig(PermissionsPresetsConfigMixin, InvenioRecordServiceConfig):
    """TestRecord service config."""
    PERMISSIONS_PRESETS = ["everyone"]
    url_prefix = "/test/"
    base_permission_policy_cls = TestPermissionPolicy
    schema = TestSchema
    search = TestSearchOptions
    record_cls = TestRecord
    service_id = "test"
    components = [ *PermissionsPresetsConfigMixin.components, *InvenioRecordServiceConfig.components]
    model = "test"
    
    @property
    def links_item(self):
        return {
            
            "self":RecordLink("{self.url_prefix}{id}"),
            
        }
    @property
    def links_search(self):
        return {
            
            **pagination_links("{self.url_prefix}{?args*}"),
            
        }
'''
    )


def test_api_views_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioAPIViewsBuilder],
        os.path.join("test", "views", "records", "api.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from flask import Blueprint

def create_api_blueprint(app):
    """Create TestRecord blueprint."""
    blueprint = app.extensions["test"].resource.as_blueprint()
    blueprint.record_once(init_create_api_blueprint)

    #calls record_once for all other functions starting with "init_addons_"
    #https://stackoverflow.com/questions/58785162/how-can-i-call-function-with-string-value-that-equals-to-function-name
    funcs = globals()
    funcs = [v for k, v in funcs.items() if k.startswith("init_addons_test") and callable(v)]
    for func in funcs:
        blueprint.record_once(func)

    return blueprint

def init_create_api_blueprint(state):
    """Init app."""
    app = state.app
    ext = app.extensions["test"]

    # register service
    sregistry = app.extensions["invenio-records-resources"].registry
    sregistry.register(ext.service, service_id="test")

    # Register indexer
    if hasattr(ext.service, "indexer"):
        iregistry = app.extensions["invenio-indexer"].registry
        iregistry.register(ext.service.indexer, indexer_id="test")
        '''
    )


def test_permissions_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordPermissionsBuilder],
        os.path.join("test", "services", "records", "permissions.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from invenio_records_permissions import RecordPermissionPolicy
# from invenio_records_permissions.generators import SystemProcess, AnyUser

class TestPermissionPolicy(RecordPermissionPolicy):
    """test.records.api.TestRecord permissions.
        Values in this class will override permission presets.
    """
    # can_search = [SystemProcess(), AnyUser()]
    # can_read = [SystemProcess(), AnyUser()]
    # can_create = [SystemProcess()]
    # can_update = [SystemProcess()]
    # can_delete = [SystemProcess()]
    # can_manage = [SystemProcess()]
'''
    )


def test_dumper_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordDumperBuilder],
        os.path.join("test", "records", "dumper.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from invenio_records.dumpers import SearchDumper

class TestDumper(SearchDumper):
    """TestRecord opensearch dumper."""
'''
    )


def test_ui_serializer_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordUISerializerBuilder],
        os.path.join("test", "resources", "records", "ui.py"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        '''
from flask_resources import BaseListSchema
from flask_resources import MarshmallowSerializer
from flask_resources.serializers import JSONSerializer

from test.services.records.ui_schema import TestUISchema

class TestUIJSONSerializer(MarshmallowSerializer):
    """UI JSON serializer."""

    def __init__(self):
        """Initialise Serializer."""
        super().__init__(
            format_serializer_cls=JSONSerializer,
            object_schema_cls=TestUISchema,
            list_schema_cls=BaseListSchema,
            schema_context={"object_key": "ui"},
        )
'''
    )



from .array import ArrayFacetsComponent
from .field import RegularFacetsComponent
from .nested import NestedFacetsComponent
from .object import ObjectFacetsComponent

__all__ = [
    "RegularFacetsComponent",
    "ObjectFacetsComponent",
    "NestedFacetsComponent",
    "ArrayFacetsComponent",
]



from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput
from oarepo_model_builder.outputs.mapping import MappingOutput
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.utils.dict import dict_get


def test_output_disabled():
    builder = ModelBuilder()
    schema = ModelSchema(
        "", {"record": {"plugins": {"output": {"disable": "__all__"}}}}
    )
    builder.set_schema(schema)
    assert (
        builder._filter_classes(
            [JSONSchemaOutput, MappingOutput],
            dict_get(schema.schema, ["record"]),
            "output",
        )
        == []
    )


def test_output_disabled_single():
    builder = ModelBuilder()
    schema = ModelSchema(
        "", {"record": {"plugins": {"output": {"disable": ["jsonschema"]}}}}
    )
    builder.set_schema(schema)
    assert set(
        x.TYPE
        for x in builder._filter_classes(
            [JSONSchemaOutput, MappingOutput],
            dict_get(schema.schema, ["record"]),
            "output",
        )
    ) == {"mapping"}


def test_output_enabled():
    builder = ModelBuilder()
    schema = ModelSchema(
        "",
        {
            "record": {
                "plugins": {"output": {"disable": "__all__", "enable": ["mapping"]}}
            }
        },
    )
    builder.set_schema(schema)
    assert set(
        x.TYPE
        for x in builder._filter_classes(
            [JSONSchemaOutput, MappingOutput],
            dict_get(schema.schema, ["record"]),
            "output",
        )
    ) == {"mapping"}


def test_output_enabled_import():
    builder = ModelBuilder()
    schema = ModelSchema(
        "",
        {
            "record": {
                "plugins": {
                    "output": {
                        "include": [
                            "oarepo_model_builder.outputs.mapping:MappingOutput"
                        ]
                    }
                }
            }
        },
    )
    builder.set_schema(schema)
    assert set(
        x.TYPE
        for x in builder._filter_classes(
            [], dict_get(schema.schema, ["record"]), "output"
        )
    ) == {"mapping"}



import marshmallow as ma
import pytest

from oarepo_model_builder.datatypes import DataType, ModelDataType, datatypes

from .multilang import UIDataTypeComponent


def test_empty():
    model = {}
    validator = ModelDataType.validator()
    assert validator.load(model) == {
        "type": "model",
        "properties": {},
        "searchable": True,
    }


def test_bad_type():
    model = {"type": "keyword"}
    validator = ModelDataType.validator()
    with pytest.raises(ma.exceptions.ValidationError):
        validator.load(model)


def test_empty_properties():
    model = {"properties": {}}
    validator = ModelDataType.validator()
    assert validator.load(model) == {
        "type": "model",
        "properties": {},
        "searchable": True,
    }


def test_bad_properties():
    model = {"properties": []}
    validator = ModelDataType.validator()
    with pytest.raises(ma.exceptions.ValidationError):
        validator.load(model)


def test_simple():
    model = {"properties": {"a": {"type": "keyword"}}}
    validator = ModelDataType.validator()
    assert validator.load(model) == {"type": "model", **model, "searchable": True}


def test_unknown_datatype():
    model = {"properties": {"a": {"type": "__unknown__"}}}
    validator = ModelDataType.validator()
    with pytest.raises(ma.exceptions.ValidationError):
        validator.load(model)


def test_object():
    model = {
        "properties": {
            "a": {"type": "object", "properties": {"prop": {"type": "keyword"}}}
        }
    }
    validator = ModelDataType.validator()
    assert validator.load(model) == {"type": "model", **model, "searchable": True}


def test_nested():
    model = {
        "properties": {
            "a": {"type": "nested", "properties": {"prop": {"type": "keyword"}}}
        }
    }
    validator = ModelDataType.validator()
    assert validator.load(model) == {"type": "model", **model, "searchable": True}


def test_array():
    model = {"properties": {"a": {"type": "array", "items": {"type": "keyword"}}}}
    validator = ModelDataType.validator()
    assert validator.load(model) == {"type": "model", **model, "searchable": True}


def test_array_of_objects():
    model = {
        "properties": {
            "a": {
                "type": "array",
                "items": {
                    "type": "nested",
                    "properties": {"prop": {"type": "keyword"}},
                },
            }
        }
    }
    validator = ModelDataType.validator()
    assert validator.load(model) == {"type": "model", **model, "searchable": True}


def test_ui():
    # remove ui component as we are providing our own
    for ci, c in reversed(list(enumerate(datatypes.components))):
        if "UI" in type(c).__name__:
            del datatypes.components[ci]

    datatypes.components.append(UIDataTypeComponent())
    assert UIDataTypeComponent.ModelSchema in datatypes.get_class_components(
        DataType, "ModelSchema"
    )
    model = {"properties": {"a": {"type": "keyword", "ui": {"test": "blah"}}}}
    validator = ModelDataType.validator()
    assert validator.load(model) == {"type": "model", **model, "searchable": True}


def test_clear_cache():
    # remove ui component as we are providing our own
    for ci, c in reversed(list(enumerate(datatypes.components))):
        if "UI" in type(c).__name__:
            del datatypes.components[ci]

    list(datatypes.components)
    datatypes.components.append(UIDataTypeComponent())
    datatypes.call_class_components(DataType, "test")
    datatypes._clear_caches()
    assert UIDataTypeComponent not in [type(x) for x in datatypes.components]



from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.schema import ModelSchema


def test_empty_builder():
    builder = ModelBuilder(output_builders=[], outputs=[])
    builder.skip_schema_validation = True
    outputs = builder.build(
        ModelSchema("", {"a": 1}, validate=False), "record", ["record"], "/tmp/test"
    )
    assert outputs == {}



import marshmallow as ma
from marshmallow import fields

from oarepo_model_builder.datatypes import datatypes
from oarepo_model_builder.utils.facet_helpers import facet_name
from oarepo_model_builder.validation.utils import ImportSchema

from ...datatypes import DataTypeComponent


class FacetsSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    searchable = ma.fields.Bool(required=False)
    key = fields.String(required=False)
    field = fields.String(required=False)
    facet_class = fields.String(required=False, data_key="facet-class")
    args = fields.List(fields.String(), required=False)
    imports = fields.List(fields.Nested(ImportSchema), required=False)
    path = fields.String(required=False)
    keyword = fields.String(required=False)


class RegularFacetsComponent(DataTypeComponent):
    eligible_datatypes = []

    class ModelSchema(ma.Schema):
        facets = ma.fields.Nested(
            FacetsSchema,
            required=False,
        )

    def build_facets(self, datatype, facets, facet_definition=None, searchable=True):
        facet_section = datatype.section_facets
        if not facet_section.config == {}:
            _path = datatype.path
            if "keyword" in datatype.section_facets.config:
                _path = _path + ".keyword"
            if "path" in facet_section.config:
                path = _path + "." + facet_section.config["path"]

            else:
                path = _path
            key = facet_section.config.get("key", facet_name(_path))
            label = facet_section.config.get(
                "label", f'{_path.replace(".", "/")}.label'
            )
            facet_searchable = facet_section.config.get("searchable", searchable)
            imports = facet_section.config.get("imports", [])
            arguments = facet_section.config.get("args", [])
            arguments_string = ",".join(arguments)
            if arguments_string != "":
                arguments_string = "," + arguments_string
            field = facet_section.config.get(
                "field",
                datatype.section_facets.config["facet_class"]
                + f'(field="{path}", label =_("{label}") {arguments_string} '
                + ")",
            )
            facet_definition = {
                "path": key,
                "class": field,
                "facet_searchable": facet_searchable,
                "imports": imports,
            }
        datatypes.call_components(
            datatype.parent,
            "build_definition",
            facets=facets,
            facet_definition=facet_definition,
            searchable=searchable,
        )

    def create_definition(self, config, datatype):
        return {
            datatype.path: config["facet_class"] + f'(field="{datatype.path}"' + ")"
        }



from oarepo_model_builder.templates import templates


def test_load_default_template():
    settings = {
        "python": {
            "templates": {},
        }
    }
    assert templates.get_template("record", settings)



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType, datatypes
from oarepo_model_builder.datatypes.components.model.utils import set_default

from .defaults import DefaultsModelComponent


class FacetsSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    searchable = ma.fields.Bool(required=False)

    extra_code = ma.fields.String(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be pasted to search options module"},
    )
    module = ma.fields.String(
        metadata={"doc": "Module where the facets will be placed"}
    )
    generate = ma.fields.Boolean()
    skip = ma.fields.Boolean()


class FacetsModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        facets = ma.fields.Nested(
            FacetsSchema,
            required=False,
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        module = datatype.definition["module"]["qualified"]
        profile_module = context["profile_module"]

        facets = set_default(datatype, "facets", {})
        facets.setdefault("generate", True)
        facets.setdefault("module", f"{module}.services.{profile_module}.facets")

        facets.setdefault("extra-code", "")

    def process_facets(self, datatype, section, **kwargs):
        facets = []
        searchable = datatype.definition.get("searchable", True)
        datatypes.call_components(
            datatype,
            "build_facets",
            facets=facets,
            facet_definition=None,
            searchable=searchable,
        )
        section.config["facets"] = facets
        datatype.definition["config"]["facets"] = facets
        return section

    def build_facets(self, datatype, facets, facet_definition=None, searchable=True):
        for c in datatype.children.values():
            if c.model_type == "array":
                children = c.item.children
            else:
                children = c.children
            if children != {}:
                self.get_leaf(children, facets, searchable)
            else:
                datatypes.call_components(
                    c,
                    "build_facets",
                    facets=facets,
                    facet_definition=None,
                    searchable=searchable,
                )

        return facets

    def get_leaf(self, children, facets, searchable):
        for c in children.values():
            if c.model_type == "array":
                children = c.item.children
            else:
                children = c.children
            if children != {}:
                self.get_leaf(children, facets, searchable)
            else:
                datatypes.call_components(
                    c,
                    "build_facets",
                    facets=facets,
                    facet_definition=None,
                    searchable=searchable,
                )

    def build_definition(
        self, datatype, facets, facet_definition=None, searchable=True
    ):
        if facet_definition:
            facet_searchable = facet_definition.get("facet_searchable", searchable)
            if facet_searchable:
                facets.append(facet_definition)



from oarepo_model_builder.datatypes import ObjectDataType, datatypes

from .field import RegularFacetsComponent


class ObjectFacetsComponent(RegularFacetsComponent):
    eligible_datatypes = [ObjectDataType]

    def build_definition(
        self, datatype, facets, facet_definition=None, searchable=True
    ):
        if datatype.key is not None:
            facet_section = datatype.section_facets.config
            if facet_section != {}:
                facet_searchable = facet_section.get("searchable", "")
                if facet_definition:
                    imports = facet_section.get("imports", [])
                    facet_definition["imports"].extend(imports)
                    fs = facet_definition.get("facet_searchable", "")
                    if fs != "" and facet_searchable != "" and fs != facet_searchable:
                        facet_definition["facet_searchable"] = facet_searchable
        datatypes.call_components(
            datatype.parent,
            "build_definition",
            facets=facets,
            facet_definition=facet_definition,
            searchable=searchable,
        )

    def build_facets(self, datatype, facets, facet_definition=None, searchable=True):
        pass



import json
import os

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.entrypoints import load_model
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.invenio.invenio_record_marshmallow import (
    InvenioRecordMarshmallowBuilder,
)
from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput
from oarepo_model_builder.outputs.mapping import MappingOutput
from oarepo_model_builder.outputs.python import PythonOutput

from .utils import strip_whitespaces


def test_enum():
    schema = load_model(
        "test.yaml",
        model_content={
            "record": {
                "use": "invenio",
                "properties": {"a": {"type": "keyword", "enum": ["a", "b", "c"]}},
                "module": {"qualified": "test"},
            },
        },
        isort=False,
        black=False,
        autoflake=False,
    )

    filesystem = InMemoryFileSystem()
    builder = ModelBuilder(
        output_builders=[
            InvenioRecordMarshmallowBuilder,
            MappingBuilder,
            JSONSchemaBuilder,
        ],
        outputs=[PythonOutput, MappingOutput, JSONSchemaOutput],
        filesystem=filesystem,
    )

    builder.build(schema, "record", ["record"], "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ).read()
    assert (
        strip_whitespaces(
            """
class TestSchema(InvenioBaseRecordSchema):

    class Meta:
        unknown = ma.RAISE


    a = ma_fields.String(validate=[ma_validate.OneOf(['a', 'b', 'c'])])
    """
        )
        in strip_whitespaces(data)
    )

    data = builder.filesystem.read(
        os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
    )
    data = json.loads(data)
    assert data == {
        "properties": {
            "$schema": {"type": "string"},
            "a": {"type": "string"},
            "created": {"format": "date-time", "type": "string"},
            "id": {"type": "string"},
            "updated": {"format": "date-time", "type": "string"},
        },
        "type": "object",
    }

    data = builder.filesystem.read(
        os.path.join("test", "records", "mappings", "os-v2", "test", "test-1.0.0.json")
    )
    data = json.loads(data)
    assert data == {
        "mappings": {
            "properties": {
                "$schema": {"type": "keyword"},
                "a": {"type": "keyword"},
                "created": {
                    "type": "date",
                    "format": "strict_date_time||strict_date_time_no_millis||basic_date_time||basic_date_time_no_millis||basic_date||strict_date",
                },
                "id": {"type": "keyword"},
                "updated": {
                    "type": "date",
                    "format": "strict_date_time||strict_date_time_no_millis||basic_date_time||basic_date_time_no_millis||basic_date||strict_date",
                },
            }
        }
    }



import marshmallow as ma

from oarepo_model_builder.datatypes import DataType, DataTypeComponent, Section
from oarepo_model_builder.utils.deepmerge import deepmerge


class MultilangPreprocessor:
    # @process(
    #     model_builder=JSONSchemaBuilder,
    #     path="/properties/*",
    #     condition=lambda current, stack: current.type == "multilingual",
    # )
    def modify_multilang_schema(self, data, **__kwargs):
        data["type"] = "object"
        data["properties"] = {"lang": {"type": "string"}, "value": {"type": "string"}}
        return data

    # @process(
    #     model_builder=MappingBuilder,
    #     path="/properties/*",
    #     condition=lambda current, stack: current.type == "multilingual",
    # )
    def modify_multilang_mapping(self, data, stack, **kwargs):
        raise ReplaceElement(
            {
                stack.top.key: {
                    "type": "object",
                    "properties": {
                        "lang": {"type": "keyword"},
                        "value": {"type": "text"},
                    },
                },
                stack.top.key + "_cs": {"type": "text"},
            }
        )

    # @process(
    #     model_builder=InvenioRecordSchemaBuilder,
    #     path="/properties/*",
    #     condition=lambda current, stack: current.type == "multilingual",
    # )
    def modify_multilang_marshmallow(self, data, **__kwargs):
        data["type"] = "object"
        deepmerge(
            data.setdefault("marshmallow", {}),
            {
                "imports": [{"import": "tests.multilang", "alias": "multilingual"}],
                "class": "multilingual.MultilingualSchema",
                "nested": True,
            },
        )
        return data


class UIValidator(ma.Schema):
    class Meta:
        unknown = ma.INCLUDE


class UIDataTypeComponent(DataTypeComponent):
    class ModelSchema(ma.Schema):
        ui = ma.fields.Nested(UIValidator)


class MultilingualDataType(DataType):
    model_type = "multilingual"

    @property
    def section_json_schema(self):
        return Section(
            "json_schema",
            config={
                "properties": {"lang": {"type": "string"}, "value": {"type": "string"}},
                "type": "object",
            },
            children={},
            item=None,
        )



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_search_options"
    section = "search-options"
    template = "record-search-options"

    def finish(self, **extra_kwargs):
        facets = self.current_model.definition["config"]["facets"]
        search_data = []
        for f in facets:
            search_data.append({f["path"]: "facets." + f["path"]})
        extra_kwargs["search_data"] = search_data
        super().finish(**extra_kwargs)



import os

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders.setup_cfg import SetupCfgBuilder
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.invenio.invenio_ext_setup_cfg import InvenioExtSetupCfgBuilder
from oarepo_model_builder.invenio.invenio_record_jsonschemas_setup_cfg import (
    InvenioRecordJSONSchemasSetupCfgBuilder,
)
from oarepo_model_builder.invenio.invenio_record_metadata_alembic_setup_cfg import (
    InvenioRecordMetadataAlembicSetupCfgBuilder,
)
from oarepo_model_builder.invenio.invenio_record_metadata_models_setup_cfg import (
    InvenioRecordMetadataModelsSetupCfgBuilder,
)
from oarepo_model_builder.invenio.invenio_record_resource_setup_cfg import (
    InvenioRecordResourceSetupCfgBuilder,
)
from oarepo_model_builder.invenio.invenio_record_search_setup_cfg import (
    InvenioRecordSearchSetupCfgBuilder,
)
from oarepo_model_builder.outputs.cfg import CFGOutput
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.schema import ModelSchema

from .utils import strip_whitespaces


def build_python_model(model, output_builders, fn):
    builder = ModelBuilder(
        output_builders=output_builders,
        outputs=[PythonOutput, CFGOutput],
        filesystem=InMemoryFileSystem(),
    )
    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {
                        "use-isort": False,
                        "use-black": False,
                        "use-autoflake": False,
                    },
                    "opensearch": {"version": "os-v2"},
                },
                "record": {"module": {"qualified": "test"}, **model},
            },
        ),
        profile="record",
        model_path=["record"],
        output_dir="",
    )

    return builder.filesystem.read(fn)


def test_setup_cfg_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [SetupCfgBuilder],
        os.path.join("setup.cfg"),  # NOSONAR
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
[metadata]
name = test
version = 1.0.0
description = Repository model for Test
authors = 


[options]
python = >=3.9
install_requires =
    oarepo>=11,<12
    oarepo-runtime>=1.0.0
packages = find:


[options.package_data]
* = *.json, *.rst, *.md, *.json5, *.jinja2
        """
    )


def test_invenio_ext_cfg_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioExtSetupCfgBuilder],
        os.path.join("setup.cfg"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
[options.entry_points]
invenio_base.api_apps = test = test.ext:TestExt
invenio_base.apps = test = test.ext:TestExt        
        """
    )


def test_invenio_jsonschemas_cfg_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordJSONSchemasSetupCfgBuilder],
        os.path.join("setup.cfg"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
[options.entry_points]
invenio_jsonschemas.schemas = test = test.records.jsonschemas      
        """
    )


def test_invenio_model_cfg_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordMetadataModelsSetupCfgBuilder],
        os.path.join("setup.cfg"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
[options.entry_points]
invenio_db.models = test = test.records.models
    
        """
    )


def test_invenio_alembic_cfg_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordMetadataAlembicSetupCfgBuilder],
        os.path.join("setup.cfg"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
[options.entry_points]
invenio_db.alembic = test = test.records:alembic
        """
    )


def test_invenio_resource_cfg_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordResourceSetupCfgBuilder],
        os.path.join("setup.cfg"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
[options.entry_points]
invenio_base.api_blueprints = test = test.views.records.api:create_api_blueprint
invenio_base.blueprints= test= test.views.records.app:create_app_blueprint
        """
    )


def test_invenio_search_cfg_builder():
    data = build_python_model(
        {"properties": {"a": {"type": "keyword"}}},
        [InvenioRecordSearchSetupCfgBuilder],
        os.path.join("setup.cfg"),
    )

    assert strip_whitespaces(data) == strip_whitespaces(
        """
[options.entry_points]
invenio_search.mappings = test = test.records.mappings        
        """
    )



import pytest
from marshmallow.exceptions import ValidationError

from oarepo_model_builder.validation.model_validation import model_validator


def test_empty_model_validation():
    assert model_validator.validate({}) == {
        "version": "1.0.0",
        "settings": {"schema-server": "local://", "oarepo-version": "11"},  # NOSONAR
    }
    assert model_validator.validate({"record": {}}) == {
        "record": {"type": "model", "properties": {}, "searchable": True},
        "settings": {"schema-server": "local://", "oarepo-version": "11"},
        "version": "1.0.0",
    }
    with pytest.raises(ValidationError, match="Must be equal to model"):
        model_validator.validate({"record": {"type": "blah"}})
    assert model_validator.validate({"record": {"properties": {}}}) == {
        "record": {"type": "model", "properties": {}, "searchable": True},
        "settings": {"schema-server": "local://", "oarepo-version": "11"},
        "version": "1.0.0",
    }
    assert model_validator.validate(
        {"record": {"type": "model", "properties": None}}
    ) == {
        "record": {"type": "model", "properties": {}, "searchable": True},
        "settings": {"schema-server": "local://", "oarepo-version": "11"},
        "version": "1.0.0",
    }


def test_unknown_on_top_validation():
    with pytest.raises(ValidationError, match="'unexpected'.*'Unknown field.'"):
        model_validator.validate({"unexpected": True})


def test_settings_on_model():
    validated = model_validator.validate(
        {
            "record": {
                "module": {"qualified": "test"},
                "json-schema-settings": {"file": "test"},
                "mapping-settings": {"file": "test"},
                "model-name": "test",
                "resource-config": {"base-url": "test"},
            }
        }
    )
    assert validated == {
        "record": {
            "json-schema-settings": {"file": "test"},
            "mapping-settings": {"file": "test"},
            "model-name": "test",
            "module": {"qualified": "test"},
            "properties": {},
            "resource-config": {"base-url": "test"},
            "type": "model",
            "searchable": True,
        },
        "settings": {"schema-server": "local://", "oarepo-version": "11"},
        "version": "1.0.0",
    }


def test_inline_props_on_model():
    validated = model_validator.validate(
        {
            "record": {
                "properties": {
                    "a": "boolean",
                    "b": "integer{minimum:1}",  # NOSONAR
                    "c": "float{exclusiveMaximum: 1.0}",  # NOSONAR
                    "d": "double",
                }
            }
        }
    )
    assert validated == {
        "version": "1.0.0",
        "record": {
            "type": "model",
            "properties": {
                "a": {"type": "boolean"},
                "b": {"type": "integer", "minimum": 1},
                "c": {"type": "float", "exclusiveMaximum": 1.0},
                "d": {"type": "double"},
            },
            "searchable": True,
        },
        "settings": {"schema-server": "local://", "oarepo-version": "11"},
    }


def test_validate_defs():
    validation_result = model_validator.validate(
        {
            "record": {},
            "$defs": {
                "a": "boolean",
                "b": "integer{minimum:1}",
                "c": "float{exclusiveMaximum: 1.0}",
                "d": "double",
            },
        }
    )
    assert validation_result == {
        "version": "1.0.0",
        "$defs": {
            "a": "boolean",
            "b": "integer{minimum:1}",
            "c": "float{exclusiveMaximum: 1.0}",
            "d": "double",
        },
        "record": {"properties": {}, "type": "model", "searchable": True},
        "settings": {"schema-server": "local://", "oarepo-version": "11"},
    }


def test_settings():
    validation_result = model_validator.validate(
        {"record": {}, "settings": {"python": {"use-black": True, "use-isort": True}}}
    )
    assert validation_result == {
        "version": "1.0.0",
        "settings": {
            "python": {"use-black": True, "use-isort": True, "use-autoflake": True},
            "schema-server": "local://",
            "oarepo-version": "11",
        },
        "record": {"properties": {}, "type": "model", "searchable": True},
    }




from oarepo_model_builder.datatypes import DataType

from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordSearchFacetsBuilder(InvenioBaseClassPythonBuilder):
    # TYPE = "invenio_record_search"
    # class_config = "record-search-options-class"
    template = "record-search-options"

    def build_node(self, node: DataType):
        # everything is done in finish
        pass

    def finish(self, **extra_kwargs):
        self._generate_facets(self.current_model, **extra_kwargs)

    def _generate_facets(self, node: DataType, **extra_kwargs):
        facets = node.definition["config"]["facets"]

        package = node.definition["facets"]["module"]
        search_options_data = []
        python_path = self.class_to_path(f"{package}.facets")
        imports = []
        for f in facets:
            search_options_data.append({f["path"]: f["class"]})
            if "imports" in f:
                for i in f["imports"]:
                    imports.append((i["import"]).rsplit(".", 1))

        imports = list(map(list, set(map(tuple, imports))))
        imports.sort(key=lambda x: (x[0], x[1]))

        self.process_template(
            python_path,
            "record-facets",
            current_package_name=package,
            search_options_data=search_options_data,
            imports=imports,
            **extra_kwargs,
        )



import os

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.datatypes import datatypes
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.schema import ModelSchema
from tests.multilang import MultilingualDataType, UIDataTypeComponent

try:
    import json5
except ImportError:
    import json as json5


def test_simple_jsonschema_builder():
    data = build({"properties": {"a": {"type": "keyword", "ui": {"class": "bolder"}}}})

    assert data == {"type": "object", "properties": {"a": {"type": "string"}}}


def test_required():
    data = build(
        {
            "properties": {
                "a": {
                    "type": "keyword",
                    "required": True,
                    "ui": {"class": "bolder"},
                }
            }
        }
    )

    assert data == {
        "type": "object",
        "properties": {"a": {"type": "string"}},
    }


def test_required_inside_metadata():
    data = build(
        {
            "properties": {
                "metadata": {
                    "marshmallow": {},  # just for debugging
                    "properties": {
                        "a": {
                            "type": "keyword",
                            "required": True,
                            "ui": {"class": "bolder"},
                        }
                    },
                }
            },
        }
    )

    assert data == {
        "type": "object",
        "properties": {
            "metadata": {
                "type": "object",
                "properties": {"a": {"type": "string"}},
            }
        },
    }


def test_min_length():
    data = build({"properties": {"a": {"type": "keyword", "minLength": 5}}})
    assert data == {
        "type": "object",
        "properties": {"a": {"type": "string"}},
    }


def test_jsonschema_preprocessor():
    data = build(
        {"properties": {"a": {"type": "multilingual", "ui": {"class": "bolder"}}}},
    )

    assert data == {
        "type": "object",
        "properties": {
            "a": {
                "type": "object",
                "properties": {
                    "lang": {"type": "string"},
                    "value": {"type": "string"},
                },
            }
        },
    }


def build(model):
    datatypes.datatype_map["multilingual"] = MultilingualDataType

    # remove ui component as we are providing our own
    for ci, c in reversed(list(enumerate(datatypes.components))):
        if "UI" in type(c).__name__:
            del datatypes.components[ci]

    datatypes.components.append(UIDataTypeComponent())
    builder = ModelBuilder(
        output_builders=[JSONSchemaBuilder],
        outputs=[JSONSchemaOutput, PythonOutput],
        filesystem=InMemoryFileSystem(),
    )
    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {
                        "use-isort": False,
                        "use-black": False,
                        "use-autoflake": False,
                    },
                },
                "record": {
                    "module": {"qualified": "test"},
                    **model,
                },
            },
        ),
        profile="record",
        model_path=["record"],
        output_dir="",
    )
    data = json5.load(
        builder.filesystem.open(
            os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
        )
    )
    return data



import json
import os

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem

from .utils import strip_whitespaces


def test_raw_type():
    schema = load_model(
        "test.yaml",
        model_content={
            "record": {
                "use": "invenio",
                "properties": {"a": {"type": "flattened"}},
                "module": {"qualified": "test"},
            },
        },
        isort=False,
        black=False,
        autoflake=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "record", ["record"], "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ).read()
    assert (
        strip_whitespaces(
            """
from invenio_records_resources.services.records.schema import BaseRecordSchema as InvenioBaseRecordSchema
from marshmallow import ValidationError
from marshmallow import validate as ma_validate
import marshmallow as ma
from marshmallow import fields as ma_fields
from marshmallow_utils import fields as mu_fields
from marshmallow_utils import schemas as mu_schemas
class TestSchema(InvenioBaseRecordSchema):
    class Meta:
        unknown = ma.RAISE
    a = ma_fields.Raw()
    """
        )
        in strip_whitespaces(data)
    )

    data = builder.filesystem.read(
        os.path.join("test", "records", "mappings", "os-v2", "test", "test-1.0.0.json")
    )
    data = json.loads(data)

    assert data == {
        "mappings": {
            "properties": {
                "a": {"type": "object", "enabled": False},
                "id": {"type": "keyword"},
                "created": {
                    "type": "date",
                    "format": "strict_date_time||strict_date_time_no_millis||basic_date_time||basic_date_time_no_millis||basic_date||strict_date",
                },
                "updated": {
                    "type": "date",
                    "format": "strict_date_time||strict_date_time_no_millis||basic_date_time||basic_date_time_no_millis||basic_date||strict_date",
                },
                "$schema": {"type": "keyword"},
            },
        }
    }

    data = builder.filesystem.read(
        os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
    )
    data = json.loads(data)
    assert data == {
        "properties": {
            "a": {"type": "object"},
            "id": {"type": "string"},
            "created": {"type": "string", "format": "date-time"},
            "updated": {"type": "string", "format": "date-time"},
            "$schema": {"type": "string"},
        },
        "type": "object",
    }



from .object import ObjectDataType


class NestedDataType(ObjectDataType):
    model_type = "nested"

    mapping = {"type": "nested"}
    facets = {
        "facet_class": "NestedLabeledFacet",
        "imports": [
            {"import": "oarepo_runtime.facets.nested_facet.NestedLabeledFacet"}
        ],
    }




import os
import re

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem

DUMMY_YAML = "test.yaml"


# def test_sort():
#     schema = load_model(
#         DUMMY_YAML,
#         "test",
#         model_content={
#             "record": {
#                 "use": "invenio",
#                 "properties": {
#                     "a": {
#                         "type": "fulltext+keyword",
#                         "sortable": {"key": "a_test"},
#                     },
#                     "b": {
#                         "type": "keyword",
#                         "facets": {"field": 'TermsFacet(field="cosi")'},
#                         "sortable": {"key": "b_test", "order": "desc"},
#                     },
#                 },
#             },
#         },
#         isort=False,
#         black=False,
#     )
#
#     filesystem = InMemoryFileSystem()
#     builder = create_builder_from_entrypoints(filesystem=filesystem)
#     builder.build(schema, "")
#
#     data = builder.filesystem.open(
#         os.path.join("test", "services", "records", "search.py")
#     ).read()
#     assert re.sub(r"\s", "", data) == re.sub(
#         r"\s",
#         "",
#         """
# from invenio_records_resources.services import SearchOptions as InvenioSearchOptions
# from flask_babelex import lazy_gettext as _
# from . import facets
#
# class TestSearchOptions(InvenioSearchOptions):
#     \"""TestRecord search options.\"""
#
#     facets = {
#     'a_keyword': facets.a_keyword,
#     'b': facets.b,
#     '_id': facets._id,
#     'created': facets.created,
#     'updated': facets.updated,
#     '_schema': facets._schema,
#     }
#     sort_options = {
#         **InvenioSearchOptions.sort_options,
#     'a_test': {'fields': ['a']},
#     'b_test': {'fields': ['-b']},
#     }
#     """,
#     )


def test_search_class():
    schema = load_model(
        DUMMY_YAML,
        model_content={
            "record": {
                "use": "invenio",
                "module": {"qualified": "test"},
                "properties": {
                    "a": {"type": "fulltext+keyword"},
                    "b": {
                        "type": "keyword",
                        "facets": {"field": "cosi"},
                    },
                },
            },
        },
        isort=False,
        black=False,
        autoflake=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "record", ["record"], "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "search.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
from invenio_records_resources.services import SearchOptions as InvenioSearchOptions
from flask_babelex import lazy_gettext as _
from . import facets


class TestSearchOptions(InvenioSearchOptions):
    \"""TestRecord search options.\"""

    facets = {
        '_schema': facets._schema,
      'a_keyword': facets.a_keyword,
      'b': facets.b,
      'created': facets.created,
      '_id': facets._id,
      'updated': facets.updated,
      
    }
    sort_options = {
        **InvenioSearchOptions.sort_options,
    }
    """,
    )



import json
import os

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders.model_saver import (
    ModelRegistrationBuilder,
    ModelSaverBuilder,
)
from oarepo_model_builder.entrypoints import (
    load_entry_points_dict,
    load_included_models_from_entry_points,
)
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.outputs.cfg import CFGOutput
from oarepo_model_builder.outputs.json import JSONOutput
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.schema import ModelSchema

try:
    import json5
except ImportError:
    import json as json5


def test_model_saver():
    data = build(
        {
            "properties": {
                "a": {"type": "keyword"},
                "b": {
                    "type": "object",
                    "properties": {
                        "c": {
                            "type": "keyword",
                        }
                    },
                    "marshmallow": {"generate": True},
                },
                "metadata": {"properties": {}},
            }
        },
    )
    assert data[0]["model"] == {
        "api-blueprint": {
            "alias": "test",
            "extra_code": "",
            "function": "test.views.records.api.create_api_blueprint",
            "generate": True,
            "imports": [],
            "module": "test.views.records.api",
        },
        "config": {
            "extra_code": "",
            "generate": True,
            "imports": [],
            "module": "test.config",
        },
        "ext": {
            "alias": "test",
            "base-classes": [],
            "class": "test.ext.TestExt",
            "extra_code": "",
            "generate": True,
            "imports": [],
            "module": "test.ext",
        },
        "facets": {
            "extra-code": "",
            "generate": True,
            "module": "test.services.records.facets",
        },
        "json-schema-settings": {
            "alias": "test",
            "file": "test/records/jsonschemas/test-1.0.0.json",
            "generate": True,
            "module": "test.records.jsonschemas",
            "name": "test-1.0.0.json",
            "version": "1.0.0",
        },
        "json-serializer": {
            "base-classes": ["MarshmallowSerializer"],
            "class": "test.resources.records.ui.TestUIJSONSerializer",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "flask_resources.BaseListSchema"},
                {"import": "flask_resources.MarshmallowSerializer"},
                {"import": "flask_resources.serializers.JSONSerializer"},
            ],
            "module": "test.resources.records.ui",
        },
        "mapping-settings": {
            "alias": "test",
            "file": "test/records/mappings/os-v2/test/test-1.0.0.json",
            "generate": True,
            "index": "test-test-1.0.0",
            "module": "test.records.mappings",
        },
        "marshmallow": {
            "base-classes": ["ma.Schema"],  # NOSONAR
            "class": "test.services.records.schema.TestSchema",
            "extra-code": "",
            "generate": True,
            "module": "test.services.records.schema",  # NOSONAR
        },
        "model-name": "Test",
        "module": {
            "alias": "test",
            "base": "test",
            "base-upper": "TEST",
            "kebab-module": "test",
            "path": "test",
            "prefix": "Test",
            "prefix-snake": "test",
            "prefix-upper": "TEST",
            "qualified": "test",
            "suffix": "test",
            "suffix-snake": "test",
            "suffix-upper": "TEST",
        },
        "permissions": {
            "base-classes": ["RecordPermissionPolicy"],
            "class": "test.services.records.permissions.TestPermissionPolicy",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_permissions.RecordPermissionPolicy"}
            ],
            "module": "test.services.records.permissions",
            "presets": ["everyone"],
        },
        "pid": {
            "context-class": "PIDFieldContext",
            "field-args": ["create=True"],
            "field-class": "PIDField",
            "generate": True,
            "imports": [
                {
                    "import": "invenio_records_resources.records.systemfields.pid.PIDField"
                },
                {
                    "import": "invenio_records_resources.records.systemfields.pid.PIDFieldContext"
                },
                {"import": "invenio_pidstore.providers.recordid_v2.RecordIdProviderV2"},
            ],
            "module": "test.records.api",  # NOSONAR
            "provider-base-classes": ["RecordIdProviderV2"],
            "provider-class": "test.records.api.TestIdProvider",
            "type": "test",
        },
        "properties": {
            "a": {"type": "keyword"},
            "b": {
                "marshmallow": {
                    "class": "test.services.records.schema.BSchema",
                    "generate": True,
                },
                "properties": {"c": {"type": "keyword"}},
                "type": "object",
                "ui": {
                    "marshmallow": {
                        "class": "test.services.records.ui_schema.BUISchema",
                        "generate": True,
                    }
                },
            },
            "metadata": {
                "marshmallow": {
                    "base-classes": ["ma.Schema"],
                    "class": "test.services.records.schema.TestMetadataSchema",
                    "extra-code": "",
                    "generate": True,
                    "module": "test.services.records.schema",
                },
                "type": "object",
                "ui": {
                    "marshmallow": {
                        "base-classes": ["ma.Schema"],
                        "class": "test.services.records.ui_schema.TestMetadataUISchema",
                        "extra-code": "",
                        "generate": True,
                        "module": "test.services.records.ui_schema",  # NOSONAR
                    }
                },
            },
        },
        "proxy": {"module": "test.proxies", "generate": True},
        "record": {
            "base-classes": ["InvenioRecord"],
            "class": "test.records.api.TestRecord",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecord",
                    "import": "invenio_records_resources.records.api.Record",
                }
            ],
            "module": "test.records.api",
        },
        "record-dumper": {
            "base-classes": ["SearchDumper"],
            "class": "test.records.dumper.TestDumper",
            "extensions": [],
            "extra-code": "",
            "generate": True,
            "imports": [{"import": "invenio_records.dumpers.SearchDumper"}],
            "module": "test.records.dumper",
        },
        "record-metadata": {
            "alembic": "test.records.alembic",
            "alias": "test",
            "base-classes": ["db.Model", "RecordMetadataBase"],
            "class": "test.records.models.TestMetadata",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records.models.RecordMetadataBase"},
                {"import": "invenio_db.db"},
            ],
            "module": "test.records.models",
            "table": "test_metadata",
        },
        "resource": {
            "base-classes": ["RecordResource"],
            "class": "test.resources.records.resource.TestResource",
            "config-key": "TEST_RECORD_RESOURCE_CLASS",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_resources.resources.RecordResource"}
            ],
            "module": "test.resources.records.resource",
            "proxy": "current_resource",
        },
        "resource-config": {
            "base-classes": ["RecordResourceConfig"],
            "base-url": "/test/",
            "class": "test.resources.records.config.TestResourceConfig",
            "config-key": "TEST_RECORD_RESOURCE_CONFIG",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_resources.resources.RecordResourceConfig"}
            ],
            "module": "test.resources.records.config",
        },
        "sample": {"file": "data/sample_data.yaml"},
        "saved-model": {
            "alias": "test",
            "file": "test/models/model.json",
            "module": "test.models",
        },
        "search-options": {
            "base-classes": ["InvenioSearchOptions"],
            "class": "test.services.records.search.TestSearchOptions",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioSearchOptions",
                    "import": "invenio_records_resources.services.SearchOptions",
                }
            ],
            "module": "test.services.records.search",
        },
        "searchable": True,
        "service": {
            "base-classes": ["InvenioRecordService"],
            "class": "test.services.records.service.TestService",
            "config-key": "TEST_RECORD_SERVICE_CLASS",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecordService",
                    "import": "invenio_records_resources.services.RecordService",
                }
            ],
            "module": "test.services.records.service",
            "proxy": "current_service",
        },
        "service-config": {
            "base-classes": [
                "PermissionsPresetsConfigMixin",
                "InvenioRecordServiceConfig",
            ],
            "class": "test.services.records.config.TestServiceConfig",
            "components": [],
            "config-key": "TEST_RECORD_SERVICE_CONFIG",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecordServiceConfig",
                    "import": "invenio_records_resources.services.RecordServiceConfig",
                },
                {
                    "import": "oarepo_runtime.config.service.PermissionsPresetsConfigMixin"
                },
            ],
            "module": "test.services.records.config",
            "service-id": "test",
        },
        "type": "model",
        "ui": {
            "marshmallow": {
                "base-classes": ["InvenioUISchema"],
                "class": "test.services.records.ui_schema.TestUISchema",
                "extra-code": "",
                "generate": True,
                "imports": [
                    {"import": "oarepo_runtime.ui.marshmallow.InvenioUISchema"}
                ],
                "module": "test.services.records.ui_schema",
            }
        },
        "app-blueprint": {
            "alias": "test",
            "extra_code": "",
            "function": "test.views.records.app.create_app_blueprint",
            "generate": True,
            "imports": [],
            "module": "test.views.records.app",
        },
    }

    assert data[1].strip() == ""
    assert (
        data[2].strip()
        == """[options.entry_points]
oarepo.models = test = test.models:model.json"""
    )


def build(model, output_builder_components=None):
    builder = ModelBuilder(
        output_builders=[
            ModelSaverBuilder,
            ModelRegistrationBuilder,
        ],
        outputs=[JSONOutput, PythonOutput, CFGOutput],
        filesystem=InMemoryFileSystem(),
    )
    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {
                        "use-isort": False,
                        "use-black": False,
                        "use-autoflake": False,
                    },
                },
                "record": {**model, "module": {"qualified": "test"}},
            },
            included_models=load_included_models_from_entry_points(),
            loaders=load_entry_points_dict("oarepo_model_builder.loaders"),
        ),
        profile="record",
        model_path=["record"],
        output_dir="",
    )
    return (
        json5.load(
            builder.filesystem.open(os.path.join("test", "models", "model.json"))
        ),
        builder.filesystem.read(os.path.join("test", "models", "__init__.py")),
        builder.filesystem.read("setup.cfg"),
    )


def test_model_saver_invenio():
    data = build(
        {"use": "invenio", "properties": {}},
    )
    print(repr(data[0]))
    assert data[0]["model"] == {
        "api-blueprint": {
            "alias": "test",
            "extra_code": "",
            "function": "test.views.records.api.create_api_blueprint",
            "generate": True,
            "imports": [],
            "module": "test.views.records.api",
        },
        "config": {
            "extra_code": "",
            "generate": True,
            "imports": [],
            "module": "test.config",
        },
        "ext": {
            "alias": "test",
            "base-classes": [],
            "class": "test.ext.TestExt",
            "extra_code": "",
            "generate": True,
            "imports": [],
            "module": "test.ext",
        },
        "facets": {
            "extra-code": "",
            "generate": True,
            "module": "test.services.records.facets",
        },
        "json-schema-settings": {
            "alias": "test",
            "file": "test/records/jsonschemas/test-1.0.0.json",
            "generate": True,
            "module": "test.records.jsonschemas",
            "name": "test-1.0.0.json",
            "version": "1.0.0",
        },
        "json-serializer": {
            "base-classes": ["MarshmallowSerializer"],
            "class": "test.resources.records.ui.TestUIJSONSerializer",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "flask_resources.BaseListSchema"},
                {"import": "flask_resources.MarshmallowSerializer"},
                {"import": "flask_resources.serializers.JSONSerializer"},
            ],
            "module": "test.resources.records.ui",
        },
        "mapping-settings": {
            "alias": "test",
            "file": "test/records/mappings/os-v2/test/test-1.0.0.json",
            "generate": True,
            "index": "test-test-1.0.0",
            "module": "test.records.mappings",
        },
        "marshmallow": {
            "base-classes": ["InvenioBaseRecordSchema"],
            "class": "test.services.records.schema.TestSchema",
            "extra-code": "",
            "generate": True,
            "module": "test.services.records.schema",
        },
        "model-name": "Test",
        "module": {
            "alias": "test",
            "base": "test",
            "base-upper": "TEST",
            "kebab-module": "test",
            "path": "test",
            "prefix": "Test",
            "prefix-snake": "test",
            "prefix-upper": "TEST",
            "qualified": "test",
            "suffix": "test",
            "suffix-snake": "test",
            "suffix-upper": "TEST",
        },
        "permissions": {
            "base-classes": ["RecordPermissionPolicy"],
            "class": "test.services.records.permissions.TestPermissionPolicy",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_permissions.RecordPermissionPolicy"}
            ],
            "module": "test.services.records.permissions",
            "presets": ["everyone"],
        },
        "pid": {
            "context-class": "PIDFieldContext",
            "field-args": ["create=True"],
            "field-class": "PIDField",
            "generate": True,
            "imports": [
                {
                    "import": "invenio_records_resources.records.systemfields.pid.PIDField"
                },
                {
                    "import": "invenio_records_resources.records.systemfields.pid.PIDFieldContext"
                },
                {"import": "invenio_pidstore.providers.recordid_v2.RecordIdProviderV2"},
            ],
            "module": "test.records.api",
            "provider-base-classes": ["RecordIdProviderV2"],
            "provider-class": "test.records.api.TestIdProvider",
            "type": "test",
        },
        "properties": {
            "$schema": {
                "facets": {"searchable": True},
                "marshmallow": {"read": False, "write": False},
                "sample": {"skip": True},
                "type": "keyword",
                "ui": {"marshmallow": {"read": False, "write": False}},
            },
            "created": {
                "facets": {"searchable": True},
                "marshmallow": {"read": False, "write": False},
                "sample": {"skip": True},
                "type": "datetime",
                "ui": {"marshmallow": {"read": False, "write": False}},
            },
            "id": {
                "facets": {"searchable": True},
                "marshmallow": {"read": False, "write": False},
                "sample": {"skip": True},
                "type": "keyword",
                "ui": {"marshmallow": {"read": False, "write": False}},
            },
            "updated": {
                "facets": {"searchable": True},
                "marshmallow": {"read": False, "write": False},
                "sample": {"skip": True},
                "type": "datetime",
                "ui": {"marshmallow": {"read": False, "write": False}},
            },
        },
        "proxy": {"module": "test.proxies", "generate": True},
        "record": {
            "base-classes": ["InvenioRecord"],
            "class": "test.records.api.TestRecord",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecord",
                    "import": "invenio_records_resources.records.api.Record",
                }
            ],
            "module": "test.records.api",
        },
        "record-dumper": {
            "base-classes": ["SearchDumper"],
            "class": "test.records.dumper.TestDumper",
            "extensions": [],
            "extra-code": "",
            "generate": True,
            "imports": [{"import": "invenio_records.dumpers.SearchDumper"}],
            "module": "test.records.dumper",
        },
        "record-metadata": {
            "alembic": "test.records.alembic",
            "alias": "test",
            "base-classes": ["db.Model", "RecordMetadataBase"],
            "class": "test.records.models.TestMetadata",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records.models.RecordMetadataBase"},
                {"import": "invenio_db.db"},
            ],
            "module": "test.records.models",
            "table": "test_metadata",
        },
        "resource": {
            "base-classes": ["RecordResource"],
            "class": "test.resources.records.resource.TestResource",
            "config-key": "TEST_RECORD_RESOURCE_CLASS",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_resources.resources.RecordResource"}
            ],
            "module": "test.resources.records.resource",
            "proxy": "current_resource",
        },
        "resource-config": {
            "base-classes": ["RecordResourceConfig"],
            "base-url": "/test/",
            "class": "test.resources.records.config.TestResourceConfig",
            "config-key": "TEST_RECORD_RESOURCE_CONFIG",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_resources.resources.RecordResourceConfig"}
            ],
            "module": "test.resources.records.config",
        },
        "sample": {"file": "data/sample_data.yaml"},
        "saved-model": {
            "alias": "test",
            "file": "test/models/model.json",
            "module": "test.models",
        },
        "search-options": {
            "base-classes": ["InvenioSearchOptions"],
            "class": "test.services.records.search.TestSearchOptions",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioSearchOptions",
                    "import": "invenio_records_resources.services.SearchOptions",
                }
            ],
            "module": "test.services.records.search",
        },
        "searchable": True,
        "service": {
            "base-classes": ["InvenioRecordService"],
            "class": "test.services.records.service.TestService",
            "config-key": "TEST_RECORD_SERVICE_CLASS",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecordService",
                    "import": "invenio_records_resources.services.RecordService",
                }
            ],
            "module": "test.services.records.service",
            "proxy": "current_service",
        },
        "service-config": {
            "base-classes": [
                "PermissionsPresetsConfigMixin",
                "InvenioRecordServiceConfig",
            ],
            "class": "test.services.records.config.TestServiceConfig",
            "components": [],
            "config-key": "TEST_RECORD_SERVICE_CONFIG",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecordServiceConfig",
                    "import": "invenio_records_resources.services.RecordServiceConfig",
                },
                {
                    "import": "oarepo_runtime.config.service.PermissionsPresetsConfigMixin"
                },
            ],
            "module": "test.services.records.config",
            "service-id": "test",
        },
        "type": "model",
        "ui": {
            "marshmallow": {
                "base-classes": ["InvenioUISchema"],
                "class": "test.services.records.ui_schema.TestUISchema",
                "extra-code": "",
                "generate": True,
                "imports": [
                    {"import": "oarepo_runtime.ui.marshmallow.InvenioUISchema"}
                ],
                "module": "test.services.records.ui_schema",
            }
        },
        "app-blueprint": {
            "alias": "test",
            "extra_code": "",
            "function": "test.views.records.app.create_app_blueprint",
            "generate": True,
            "imports": [],
            "module": "test.views.records.app",
        },
    }



from oarepo_model_builder.schema import ModelSchema


def test_prepare_datatype():
    model = ModelSchema(
        "",
        {
            "settings": {
                "python": {
                    "use-isort": False,
                    "use-black": False,
                    "use-autoflake": False,
                },
                "opensearch": {"version": "os-v2"},
            },
            "record": {
                "type": "model",
                "model-name": "My Test Record",
                "module": {"qualified": "my.test"},
                "properties": {"metadata": {"type": "object"}},
            },
        },
    )
    record_section = model.get_schema_section(profile="record", section="record")
    assert record_section.definition == {
        "api-blueprint": {
            "alias": "my_test_record",
            "extra_code": "",
            "function": "my.test.views.records.api.create_api_blueprint",
            "generate": True,
            "imports": [],
            "module": "my.test.views.records.api",
        },
        "app-blueprint": {  # NOSONAR
            "alias": "my_test_record",
            "extra_code": "",
            "function": "my.test.views.records.app.create_app_blueprint",
            "generate": True,
            "imports": [],
            "module": "my.test.views.records.app",
        },
        "config": {
            "extra_code": "",
            "generate": True,
            "imports": [],
            "module": "my.test.config",
        },
        "ext": {
            "alias": "my_test_record",
            "base-classes": [],
            "class": "my.test.ext.TestExt",
            "extra_code": "",
            "generate": True,
            "imports": [],
            "module": "my.test.ext",
        },
        "facets": {
            "extra-code": "",
            "generate": True,
            "module": "my.test.services.records.facets",
        },
        "json-schema-settings": {
            "alias": "my_test_record",
            "file": "my/test/records/jsonschemas/test-1.0.0.json",
            "generate": True,
            "module": "my.test.records.jsonschemas",
            "name": "test-1.0.0.json",
            "version": "1.0.0",
        },
        "json-serializer": {
            "base-classes": ["MarshmallowSerializer"],
            "class": "my.test.resources.records.ui.TestUIJSONSerializer",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "flask_resources.BaseListSchema"},
                {"import": "flask_resources.MarshmallowSerializer"},
                {"import": "flask_resources.serializers.JSONSerializer"},
            ],
            "module": "my.test.resources.records.ui",
        },
        "mapping-settings": {
            "alias": "my_test_record",
            "file": "my/test/records/mappings/os-v2/my_test_record/test-1.0.0.json",
            "generate": True,
            "index": "my_test_record-test-1.0.0",
            "module": "my.test.records.mappings",
        },
        "marshmallow": {
            "base-classes": ["ma.Schema"],  # NOSONAR
            "class": "my.test.services.records.schema.TestSchema",
            "extra-code": "",
            "generate": True,
            "module": "my.test.services.records.schema",
        },
        "model-name": "My Test Record",
        "module": {
            "alias": "my_test_record",
            "base": "test",
            "base-upper": "TEST",
            "kebab-module": "my-test",
            "path": "my/test",
            "prefix": "Test",
            "prefix-snake": "test",
            "prefix-upper": "TEST",
            "qualified": "my.test",
            "suffix": "test",
            "suffix-snake": "test",
            "suffix-upper": "TEST",
        },
        "permissions": {
            "base-classes": ["RecordPermissionPolicy"],
            "class": "my.test.services.records.permissions.TestPermissionPolicy",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_permissions.RecordPermissionPolicy"}
            ],
            "module": "my.test.services.records.permissions",
            "presets": ["everyone"],
        },
        "pid": {
            "context-class": "PIDFieldContext",
            "field-args": ["create=True"],
            "field-class": "PIDField",
            "generate": True,
            "imports": [
                {
                    "import": "invenio_records_resources.records.systemfields.pid.PIDField"
                },
                {
                    "import": "invenio_records_resources.records.systemfields.pid.PIDFieldContext"
                },
                {"import": "invenio_pidstore.providers.recordid_v2.RecordIdProviderV2"},
            ],
            "module": "my.test.records.api",
            "provider-base-classes": ["RecordIdProviderV2"],
            "provider-class": "my.test.records.api.TestIdProvider",
            "type": "mytcrd",
        },
        "properties": {
            "metadata": {
                "marshmallow": {
                    "base-classes": ["ma.Schema"],
                    "class": "my.test.services.records.schema.TestMetadataSchema",
                    "extra-code": "",
                    "generate": True,
                    "module": "my.test.services.records.schema",
                },
                "properties": {},
                "type": "object",
                "ui": {
                    "marshmallow": {
                        "base-classes": ["ma.Schema"],
                        "class": "my.test.services.records.ui_schema.TestMetadataUISchema",
                        "extra-code": "",
                        "generate": True,
                        "module": "my.test.services.records.ui_schema",
                    }
                },
            }
        },
        "proxy": {"module": "my.test.proxies", "generate": True},
        "record": {
            "base-classes": ["InvenioRecord"],
            "class": "my.test.records.api.TestRecord",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecord",
                    "import": "invenio_records_resources.records.api.Record",
                }
            ],
            "module": "my.test.records.api",
        },
        "record-dumper": {
            "base-classes": ["SearchDumper"],
            "class": "my.test.records.dumper.TestDumper",
            "extensions": [],
            "extra-code": "",
            "generate": True,
            "imports": [{"import": "invenio_records.dumpers.SearchDumper"}],
            "module": "my.test.records.dumper",
        },
        "record-metadata": {
            "alembic": "my.test.records.alembic",
            "alias": "my_test_record",
            "base-classes": ["db.Model", "RecordMetadataBase"],
            "class": "my.test.records.models.TestMetadata",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records.models.RecordMetadataBase"},
                {"import": "invenio_db.db"},
            ],
            "module": "my.test.records.models",
            "table": "test_metadata",
        },
        "resource": {
            "base-classes": ["RecordResource"],
            "class": "my.test.resources.records.resource.TestResource",
            "config-key": "TEST_RECORD_RESOURCE_CLASS",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_resources.resources.RecordResource"}
            ],
            "module": "my.test.resources.records.resource",
            "proxy": "current_resource",
        },
        "resource-config": {
            "base-classes": ["RecordResourceConfig"],
            "base-url": "/my-test/",
            "class": "my.test.resources.records.config.TestResourceConfig",
            "config-key": "TEST_RECORD_RESOURCE_CONFIG",
            "extra-code": "",
            "generate": True,
            "imports": [
                {"import": "invenio_records_resources.resources.RecordResourceConfig"}
            ],
            "module": "my.test.resources.records.config",
        },
        "sample": {"file": "data/sample_data.yaml"},
        "saved-model": {
            "alias": "my_test_record",
            "file": "my/test/models/model.json",
            "module": "my.test.models",
        },
        "search-options": {
            "base-classes": ["InvenioSearchOptions"],
            "class": "my.test.services.records.search.TestSearchOptions",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioSearchOptions",
                    "import": "invenio_records_resources.services.SearchOptions",
                }
            ],
            "module": "my.test.services.records.search",
        },
        "searchable": True,
        "service": {
            "base-classes": ["InvenioRecordService"],
            "class": "my.test.services.records.service.TestService",
            "config-key": "TEST_RECORD_SERVICE_CLASS",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecordService",
                    "import": "invenio_records_resources.services.RecordService",
                }
            ],
            "module": "my.test.services.records.service",
            "proxy": "current_service",
        },
        "service-config": {
            "base-classes": [
                "PermissionsPresetsConfigMixin",
                "InvenioRecordServiceConfig",
            ],
            "class": "my.test.services.records.config.TestServiceConfig",
            "components": [],
            "config-key": "TEST_RECORD_SERVICE_CONFIG",
            "extra-code": "",
            "generate": True,
            "imports": [
                {
                    "alias": "InvenioRecordServiceConfig",
                    "import": "invenio_records_resources.services.RecordServiceConfig",
                },
                {
                    "import": "oarepo_runtime.config.service.PermissionsPresetsConfigMixin"
                },
            ],
            "module": "my.test.services.records.config",
            "service-id": "my_test_record",
        },
        "type": "model",
        "ui": {
            "marshmallow": {
                "base-classes": ["InvenioUISchema"],
                "class": "my.test.services.records.ui_schema.TestUISchema",
                "extra-code": "",
                "generate": True,
                "imports": [
                    {"import": "oarepo_runtime.ui.marshmallow.InvenioUISchema"}
                ],
                "module": "my.test.services.records.ui_schema",
            }
        },
        "app-blueprint": {
            "alias": "my_test_record",
            "extra_code": "",
            "function": "my.test.views.records.app.create_app_blueprint",
            "generate": True,
            "imports": [],
            "module": "my.test.views.records.app",
        },
    }


def test_ids():
    model = ModelSchema(
        "",
        {
            "settings": {
                "python": {
                    "use-isort": False,
                    "use-black": False,
                    "use-autoflake": False,
                },
                "opensearch": {"version": "os-v2"},
            },
            "record": {
                "type": "model",
                "properties": {
                    "a": {"type": "integer", "id": "_a"},
                    "b": {
                        "type": "object",
                        "properties": {
                            "c": {"type": "integer", "id": "_c"},
                            "d": {
                                "type": "array",
                                "id": "_d",
                                "items": {
                                    "type": "object",
                                    "id": "_di",
                                    "properties": {
                                        "e": {"type": "integer", "id": "_e"},
                                    },
                                },
                            },
                        },
                        "id": "_b",
                    },
                },
            },
        },
    )
    dt = model.get_schema_section("record", "record")
    assert dt.id is None
    assert dt.children["a"].id == "_a"
    assert dt.children["b"].id == "_b"
    assert dt.children["b"].children["c"].id == "_c"
    assert dt.children["b"].children["d"].id == "_d"
    assert dt.children["b"].children["d"].item.id == "_di"
    assert dt.children["b"].children["d"].item.children["e"].id == "_e"



from oarepo_model_builder.datatypes import ArrayDataType, datatypes
from oarepo_model_builder.utils.facet_helpers import facet_name

from .field import RegularFacetsComponent


class ArrayFacetsComponent(RegularFacetsComponent):
    eligible_datatypes = [ArrayDataType]

    def build_definition(
        self, datatype, facets, facet_definition=None, searchable=True
    ):
        facet_section = datatype.section_facets.config
        if facet_section != {}:
            facet_searchable = facet_section.get("searchable", "")
            if facet_definition:
                imports = facet_section.get("imports", [])
                facet_definition["imports"].extend(imports)
                fs = facet_definition.get("facet_searchable", "")
                if fs != "" and facet_searchable != "" and fs != facet_searchable:
                    facet_definition["facet_searchable"] = facet_searchable

        datatypes.call_components(
            datatype.parent,
            "build_definition",
            facets=facets,
            facet_definition=facet_definition,
            searchable=searchable,
        )

    def build_facets(self, datatype, facets, facet_definition=None, searchable=True):
        if hasattr(datatype.item, "facets"):
            facet_section = datatype.item.facets
            _path = datatype.path
            if "keyword" in facet_section:
                _path = _path + ".keyword"
            if "path" in facet_section:
                path = _path + "." + facet_section["path"]
            else:
                path = _path
            key = facet_section.get("key", facet_name(_path))
            label = facet_section.get("label", f'{_path.replace(".", "/")}.label')
            imports = facet_section.get("imports", [])
            arguments = facet_section.get("args", [])
            arguments_string = ",".join(arguments)
            if arguments_string != "":
                arguments_string = "," + arguments_string
            facet_searchable = facet_section.get("searchable", searchable)
            field = facet_section.get(
                "field",
                facet_section["facet_class"]
                + f'(field="{path}", label =_("{label}") {arguments_string} '
                + ")",
            )
            facet_definition = {
                "path": key,
                "class": field,
                "facet_searchable": facet_searchable,
                "imports": imports,
            }

        datatypes.call_components(
            datatype.parent,
            "build_definition",
            facets=facets,
            facet_definition=facet_definition,
            searchable=searchable,
        )



import json
import os

import yaml

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem

from .utils import strip_whitespaces


def test_array_shortcuts():
    schema = load_model(
        "test.yaml",  # NOSONAR
        model_content={
            "record": {
                "use": "invenio",
                "properties": {"a[]": {"type": "keyword", "^required": True}},
                "module": {"qualified": "test"},
            },
        },
        isort=False,
        black=False,
        autoflake=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "record", ["record"], "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ).read()
    assert (
        strip_whitespaces(
            """
class TestSchema(InvenioBaseRecordSchema):

    class Meta:
        unknown = ma.RAISE


    a = ma_fields.List(ma_fields.String())    
    """
        )
        in strip_whitespaces(data)
    )

    data = builder.filesystem.read(
        os.path.join("test", "records", "mappings", "os-v2", "test", "test-1.0.0.json")
    )
    data = json.loads(data)
    assert data == {
        "mappings": {
            "properties": {
                "$schema": {"type": "keyword"},
                "a": {"type": "keyword"},
                "created": {
                    "type": "date",
                    "format": "strict_date_time||strict_date_time_no_millis||basic_date_time||basic_date_time_no_millis||basic_date||strict_date",
                },
                "id": {"type": "keyword"},
                "updated": {
                    "type": "date",
                    "format": "strict_date_time||strict_date_time_no_millis||basic_date_time||basic_date_time_no_millis||basic_date||strict_date",
                },
            }
        }
    }

    data = builder.filesystem.read(
        os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
    )
    data = json.loads(data)
    assert data == {
        "properties": {
            "$schema": {"type": "string"},
            "a": {"items": {"type": "string"}, "type": "array"},
            "created": {"format": "date-time", "type": "string"},
            "id": {"type": "string"},
            "updated": {"format": "date-time", "type": "string"},
        },
        "type": "object",
    }


def test_singleline_type_shortcut():
    data = """
record:
  properties:
    metadata:
      properties:
        title: fulltext
  module:
    qualified: test
"""

    schema = load_model(
        "test.yaml",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
        autoflake=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "record", ["record"], "")


def test_object_shortcut():
    data = """
record:
  properties:
    metadata{}:
      title: fulltext
  module:
    qualified: test
"""

    schema = load_model(
        "test.yaml",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
        autoflake=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "record", ["record"], "")


def test_array_shortcut():
    data = """
record:
  module:
    qualified: test
  properties:
    metadata[]:
      type: fulltext
"""

    schema = load_model(
        "test.yaml",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
        autoflake=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "record", ["record"], "")


def test_array_single_line_shortcut():
    data = """
record:
  module:
    qualified: test
  properties:
    metadata[]: fulltext
"""

    schema = load_model(
        "test.yaml",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
        autoflake=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "record", ["record"], "")


def test_misc_shortcuts():
    data = """
record:
  module:
    qualified: test
  properties:
    metadata:
      properties:
        title: fulltext
"""
    """
      description: fulltext+keyword
      kind: keyword{enum:[article, dataset]}
      printed{}:
      ^marshmallow:
        schema-class: PrintedPublicationSchema
      price: double{minimumExclusive:0}
      floatPrice: float     # just to test floats
      published: date
      "sources{nested}[]":
        source: fulltext
        period: edtf-interval
"""

    schema = load_model(
        "test.yaml",
        model_content=yaml.safe_load(data),
        isort=False,
        black=False,
        autoflake=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "record", ["record"], "")



import os

import pytest

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.invenio.invenio_record_resource_config import (
    InvenioRecordResourceConfigBuilder,
)
from oarepo_model_builder.invenio.invenio_record_ui_marshmallow import (
    InvenioRecordUIMarshmallowBuilder,
)
from oarepo_model_builder.invenio.invenio_record_ui_serializer import (
    InvenioRecordUISerializerBuilder,
)
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.schema import ModelSchema

from .utils import strip_whitespaces


def get_test_schema(**props):
    return ModelSchema(
        "",
        {
            "settings": {
                "python": {
                    "use-isort": False,
                    "use-black": False,
                    "use-autoflake": False,
                },
            },
            "record": {"module": {"qualified": "test"}, "properties": props},
        },
    )


@pytest.fixture
def fulltext_builder():
    return ModelBuilder(
        output_builders=[
            InvenioRecordUIMarshmallowBuilder,
            InvenioRecordUISerializerBuilder,
            InvenioRecordResourceConfigBuilder,
        ],
        outputs=[PythonOutput],
    )


def _test(fulltext_builder, string_type):
    schema = get_test_schema(a={"type": string_type})
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")  # NOSONAR
    ) as f:
        data = f.read()
    assert "a = ma_fields.String()" in data


def test_fulltext(fulltext_builder):
    _test(fulltext_builder, "fulltext")


def test_keyword(fulltext_builder):
    _test(fulltext_builder, "keyword")


def test_fulltext_keyword(fulltext_builder):
    _test(fulltext_builder, "fulltext+keyword")


def test_simple_array(fulltext_builder):
    schema = get_test_schema(a={"type": "array", "items": {"type": "keyword"}})
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert "a = ma_fields.List(ma_fields.String())" in data


def test_array_of_objects(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "array",
            "items": {"type": "object", "properties": {"b": {"type": "integer"}}},
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert (
        strip_whitespaces(
            """ 
class TestUISchema(InvenioUISchema):

    class Meta:
        unknown = ma.RAISE


    a = ma_fields.List(ma_fields.Nested(lambda: AItemUISchema()))


class AItemUISchema(ma.Schema):

    class Meta:
        unknown = ma.RAISE


    b = ma_fields.Integer() """
        )
        in strip_whitespaces(data)
    )


def test_generate_nested_schema_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            "marshmallow": {"class": "B", "generate": True},
            "ui": {"marshmallow": {"class": "BUISchema", "generate": True}},
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert (
        strip_whitespaces(
            """
class TestUISchema(InvenioUISchema):

    class Meta:
        unknown = ma.RAISE


    a = ma_fields.Nested(lambda: BUISchema())


class BUISchema(ma.Schema):

    class Meta:
        unknown = ma.RAISE


    b = ma_fields.String()    
    """
        )
        in strip_whitespaces(data)
    )


def test_generate_nested_schema_different_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            "marshmallow": {
                "class": "test.services.schema2.B",
                "generate": True,
            },
            "ui": {
                "marshmallow": {
                    "class": "test.services.schema2.BUISchema",
                    "generate": True,
                }
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert (
        strip_whitespaces(
            """
from test.services.schema2 import BUISchema
class TestUISchema(InvenioUISchema):
    class Meta:
        unknown = ma.RAISE
    a = ma_fields.Nested(lambda: BUISchema())
    """
        )
        in strip_whitespaces(data)
    )


def test_use_nested_schema_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            "marshmallow": {"class": "B", "generate": False},
            "ui": {"marshmallow": {"class": "BUISchema", "generate": False}},
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()

    assert (
        strip_whitespaces(
            """
class TestUISchema(InvenioUISchema):

    class Meta:
        unknown = ma.RAISE


    a = ma_fields.Nested(lambda: BUISchema())
"""
        )
        in strip_whitespaces(data)
    )

    assert strip_whitespaces("class BUISchema") not in strip_whitespaces(data)


def test_use_nested_schema_different_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            "marshmallow": {"class": "c.B", "generate": False},
            "ui": {"marshmallow": {"class": "c.BUISchema", "generate": False}},
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()

    assert (
        strip_whitespaces(
            """
from c import BUISchema
from oarepo_runtime.ui.marshmallow import InvenioUISchema
class TestUISchema(InvenioUISchema):
    class Meta:
        unknown = ma.RAISE
    a = ma_fields.Nested(lambda: BUISchema())
"""
        )
        in strip_whitespaces(data)
    )


def test_generate_nested_schema_array(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "array",
            "items": {
                "type": "object",
                "marshmallow": {"class": "B", "generate": True},
                "ui": {"marshmallow": {"class": "BUISchema", "generate": True}},
                "properties": {
                    "b": {
                        "type": "keyword",
                    }
                },
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert (
        strip_whitespaces(
            """
class TestUISchema(InvenioUISchema):
    class Meta:
        unknown = ma.RAISE
    a = ma_fields.List(ma_fields.Nested(lambda: BUISchema()))


class BUISchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE
    b = ma_fields.String()
"""
        )
        in strip_whitespaces(data)
    )


def test_extend_existing(fulltext_builder):
    schema = get_test_schema(a={"type": "keyword"}, b={"type": "keyword"})
    fulltext_builder.filesystem = InMemoryFileSystem()

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py"), "w"
    ) as f:
        f.write(
            '''
from invenio_records_resources.services.records.schema import BaseRecordSchema as InvenioBaseRecordSchema
import marshmallow as ma
import marshmallow.fields as ma_fields
import marshmallow.validate as ma_valid
class TestUISchema(ma.Schema):
    """TestUISchema schema."""
    a = ma_fields.String()'''
        )

    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )
    data = fulltext_builder.filesystem.read(
        os.path.join("test", "services", "records", "ui_schema.py")
    )
    assert "b = ma_fields.String()" in data


def test_generate_nested_schema_relative_same_package(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            "marshmallow": {
                "class": ".schema2.B",
                "generate": True,
            },
            "ui": {
                "marshmallow": {
                    "class": "..ui_schema2.BUISchema",
                    "generate": True,
                }
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert (
        strip_whitespaces(
            """
class TestUISchema(InvenioUISchema):
    class Meta:
        unknown = ma.RAISE
    a = ma_fields.Nested(lambda: BUISchema())
"""
        )
        in strip_whitespaces(data)
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema2.py")
    ) as f:
        data = f.read()
    assert (
        strip_whitespaces(
            """
class BUISchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE
    b = ma_fields.String()
"""
        )
        in strip_whitespaces(data)
    )


def test_generate_nested_schema_relative_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            "marshmallow": {
                "class": ".B",
                "generate": True,
            },
            "ui": {
                "marshmallow": {
                    "class": ".BUISchema",
                    "generate": True,
                }
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()

    assert (
        strip_whitespaces(
            """
class TestUISchema(InvenioUISchema):
    class Meta:
        unknown = ma.RAISE
    a = ma_fields.Nested(lambda: BUISchema())

class BUISchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE
    b = ma_fields.String()
"""
        )
        in strip_whitespaces(data)
    )


def test_generate_nested_schema_relative_upper(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            "marshmallow": {
                "class": "...schema2.B",
                "generate": True,
            },
            "ui": {
                "marshmallow": {
                    "class": "..schema2.BUISchema",
                    "generate": True,
                }
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()

    assert (
        strip_whitespaces(
            """
from test.services.records.schema2 import BUISchema
class TestUISchema(InvenioUISchema):
    class Meta:
        unknown = ma.RAISE
    a = ma_fields.Nested(lambda: BUISchema())
"""
        )
        in strip_whitespaces(data)
    )


def test_generate_json_serializer(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "keyword",
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )
    with fulltext_builder.filesystem.open(
        os.path.join("test", "resources", "records", "ui.py")
    ) as f:
        data = f.read()
    assert (
        strip_whitespaces(
            '''
from flask_resources import BaseListSchema
from flask_resources import MarshmallowSerializer
from flask_resources.serializers import JSONSerializer

from test.services.records.ui_schema import TestUISchema



class TestUIJSONSerializer(MarshmallowSerializer):
    """UI JSON serializer."""

    def __init__(self):
        """Initialise Serializer."""
        super().__init__(
            format_serializer_cls=JSONSerializer,
            object_schema_cls=TestUISchema,
            list_schema_cls=BaseListSchema,
            schema_context={"object_key": "ui"},
        )    
    '''
        )
        == strip_whitespaces(data)
    )


def test_localized_date(fulltext_builder):
    schema = get_test_schema(a={"type": "date"})
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert "a = l10n.LocalizedDate()" in data


def test_metadata(fulltext_builder):
    schema = get_test_schema(
        metadata={"type": "object", "properties": {"a": {"type": "keyword"}}}
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(
        schema, profile="record", model_path=["record"], output_dir=""
    )

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert "metadata = ma_fields.Nested(lambda: TestMetadataUISchema())" in data



from oarepo_model_builder.datatypes import NestedDataType, datatypes

from .object import ObjectFacetsComponent


class NestedFacetsComponent(ObjectFacetsComponent):
    eligible_datatypes = [NestedDataType]

    def build_definition(
        self, datatype, facets, facet_definition=None, searchable=True
    ):
        inner_facet_definition = facet_definition["class"]
        facet_section = datatype.section_facets.config
        if facet_section != {}:
            _path = datatype.path
            if "path" in facet_section:
                path = _path + "." + facet_section["path"]
            else:
                path = _path

            facet_searchable = facet_section.get("searchable", "")
            if facet_definition:
                fs = facet_definition.get("facet_searchable", "")
                if fs != "" and facet_searchable != "" and fs != facet_searchable:
                    facet_definition["facet_searchable"] = facet_searchable
                imports = facet_section.get("imports", [])
                arguments = facet_section.get("args", [])
                arguments_string = ",".join(arguments)
                if arguments_string != "":
                    arguments_string = "," + arguments_string
                facet_definition["imports"].extend(imports)
                facet_definition["class"] = facet_section.get(
                    "field",
                    f'{datatype.section_facets.config["facet_class"]}(path = "{path}", nested_facet = {inner_facet_definition} {arguments_string})',
                )
        datatypes.call_components(
            datatype.parent,
            "build_definition",
            facets=facets,
            facet_definition=facet_definition,
            searchable=searchable,
        )

    def build_facets(self, datatype, facets, facet_definition=None, searchable=True):
        pass
